
-- # import id

mutual
  data Ctx : Set where
    nil	 : Ctx
    snoc : (Γ : Ctx) -> (A : Typ Γ) -> Ctx

  data Typ : Ctx -> Set where
    set	(Γ : Ctx) : Typ Γ
    pi  (Γ : Ctx) : (A : Typ Γ) -> (B : Typ (snoc Γ A)) -> Typ Γ

-- Eliminations

{-

Pc : Ctx -> Set
Pt : (Γ : Ctx) -> Typ Γ -> Set


		Pc Γ   Pt Γ A    	  Pc Γ	  
	 ∀ Γ A. -------------    ∀ Γ. ------------
Pc nil		Pc (snoc Γ A)         Pt Γ (set Γ)

             Pc Γ   Pt Γ A   Pt (snoc Γ A) B
    ∀ Γ A B. -------------------------------
		     Pt Γ (pi Γ A B)
--------------------------------------------------
	(∀ Γ. Pc Γ)  ∧  (∀ Γ A. Pt Γ A)

We can't make Γ a parameter to Typ for two reasons:
a) Typ would become a nested type which we can't handle in the first place.
b) Parameters have to exist before the type. Trying to write out the
   elimination rule will reveal this.

-}

-- Question: can we demutual this definition?

-- First, let's try to get rid of the mutuality in the indices.

-- We start by representing the raw terms
data Raw : Set where
  nilR	: Raw
  snocR	: Raw -> Raw -> Raw
  setR	: Raw -> Raw
  piR	: Raw -> Raw -> Raw -> Raw

-- Now we use the raw terms as indices
mutual
  data Ctx' : Raw -> Set where
    nil'  : Ctx' nilR
    snoc' : (Γ : Raw) -> Ctx' Γ -> (A : Raw) -> Typ' Γ A -> Ctx' (snocR Γ A)

  data Typ' : Raw -> Raw -> Set where
    set' : (Γ : Raw) -> Ctx' Γ -> Typ' Γ (setR Γ)
    pi'	 : (Γ : Raw) -> Ctx' Γ -> (A : Raw) -> Typ' Γ A ->
				  (B : Raw) -> Typ' (snocR Γ A) B ->
				  Typ' Γ (piR Γ A B)

-- Ok, now let's remove the mutuality altogether.
data Index : Set where
  ctxIx : Raw -> Index
  typIx : Raw -> Raw -> Index

data CtxTyp : Index -> Set where
  nilE	: CtxTyp (ctxIx nilR)
  snocE	(Γ, A : Raw) : CtxTyp (ctxIx Γ) ->
		       CtxTyp (typIx Γ A) ->
		       CtxTyp (ctxIx (snocR Γ A))
  setE	(Γ : Raw) : CtxTyp (ctxIx Γ) -> CtxTyp (typIx Γ (setR Γ))
  piE	(Γ, A, B : Raw) :
	  CtxTyp (ctxIx Γ) -> CtxTyp (typIx Γ A) ->
	  CtxTyp (typIx (snocR Γ A) B) ->
	  CtxTyp (typIx Γ (piR Γ A B))

CtxE : Raw -> Set
CtxE Γ = CtxTyp (ctxIx Γ)

TypE : Raw -> Raw -> Set
TypE Γ A = CtxTyp (typIx Γ A)

-- Let's prove that the encoding is isomorphic to the original:
mutual
  eraseCtx : Ctx -> Raw
  eraseCtx nil	      = nilR
  eraseCtx (snoc Γ A) = snocR (eraseCtx Γ) (eraseTyp A)

  eraseTyp (Γ : Ctx) : Typ Γ -> Raw
  eraseTyp {Γ} set      = setR (eraseCtx Γ)
  eraseTyp {Γ} (pi A B) = piR (eraseCtx Γ) (eraseTyp A) (eraseTyp B)

mutual
  toCtxE : (Γ : Ctx) -> CtxE (eraseCtx Γ)
  toCtxE nil	    = nilE
  toCtxE (snoc Γ A) = snocE (toCtxE Γ) (toTypE A)

  toTypE (Γ : Ctx) : (A : Typ Γ) -> TypE (eraseCtx Γ) (eraseTyp A)
  toTypE {Γ} set      = setE (toCtxE Γ)
  toTypE {Γ} (pi A B) = piE (toCtxE Γ) (toTypE A) (toTypE B)

mutual
  fromCtxE (rΓ : Raw) : CtxE rΓ -> Ctx
  fromCtxE nilE		= nil
  fromCtxE (snocE Γ A)	= snoc (fromCtxE Γ) (fromTypE Γ A)

  fromTypE (rΓ, rA : Raw) : (Γ : CtxE rΓ) ->
			    TypE rΓ rA -> Typ (fromCtxE Γ)
  fromTypE Γ (setE _)	  = set
  fromTypE Γ (piE _ A B)  = pi (fromTypE Γ A) (fromTypE (snocE Γ A) B)

mutual
  eraseFromCtx : (rΓ : Raw) -> (Γ : CtxE rΓ) -> eraseCtx (fromCtxE Γ) == rΓ
  eraseFromCtx nilR nilE = refl
  eraseFromCtx (snocR rΓ rA) (snocE Γ A) =
      rewrite eraseFromCtx rΓ Γ
	      eraseFromTyp rΓ Γ rA A
      in refl
  eraseFromTyp : (rΓ : Raw) -> (Γ : CtxE rΓ) -> (rA : Raw) -> (A : TypE rΓ rA) ->
		 eraseTyp (fromTypE Γ A) == rA
  eraseFromTyp rΓ Γ (setR rΓ)	   (setE _)    =
      rewrite eraseFromCtx rΓ in refl
  eraseFromTyp rΓ Γ (piR rΓ rA rB) (piE _ A B) =
      rewrite eraseFromCtx rΓ
	      eraseFromTyp rΓ Γ rA A
	      eraseFromTyp (snocR rΓ rA) (snocE Γ A) rB B
      in refl

mutual
  sameCtx (rΓ : Raw) : (Γ, Γ' : CtxE rΓ) -> Γ == Γ'
  sameCtx nilE nilE		    = refl
  sameCtx (snocE Γ A) (snocE Γ' A') = cong2 snocE (sameCtx Γ Γ') (sameTyp A A')

  sameTyp (rΓ, rA : Raw) : (A, A' : TypE rΓ rA) -> A == A'
  sameTyp (setE Γ)    (setE Γ')	     = cong setE (sameCtx Γ Γ')
  sameTyp (piE Γ A B) (piE Γ' A' B') = cong3 piE (sameCtx Γ Γ') (sameTyp A A')
						 (sameTyp B B')

rewrite eraseFromCtx

ToFromTypId (rΓ, rA : Raw) : (Γ : CtxE rΓ) -> (A : TypE rΓ rA) ->Set
ToFromTypId Γ A = rewrite eraseFromTyp _ Γ in
		  toTypE (fromTypE Γ A) == A

mutual
  toFromCtxId (rΓ : Raw) : (Γ : CtxE rΓ) -> toCtxE (fromCtxE Γ) == Γ
  toFromCtxId nilE	  = refl
  toFromCtxId (snocE Γ A) = rewrite toFromCtxId Γ
				    toFromTypId Γ A
				    eraseFromTyp _ Γ _ A
			    in refl

  toFromTypId (rΓ, rA : Raw) : (Γ : CtxE rΓ) -> (A : TypE rΓ rA) ->
			       ToFromTypId Γ A
  toFromTypId Γ (setE _)    = cong setE (sameCtx _ _)
  toFromTypId Γ (piE _ A B) =
    rewrite eraseFromTyp _ Γ _ A
	    eraseFromTyp _ (snocE Γ A) _ B
    in	cong3 piE (sameCtx _ _) (sameTyp _ _) (sameTyp _ _)

mutual
  FromToTypId (Γ : Ctx) : (A : Typ Γ) -> Set
  FromToTypId {Γ} A =
      rewrite fromToCtxId Γ
      in      fromTypE (toCtxE Γ) (toTypE A) == A

  fromToCtxId : (Γ : Ctx) -> fromCtxE (toCtxE Γ) == Γ
  fromToCtxId nil	 = refl
  fromToCtxId (snoc Γ A) =
    rewrite fromToCtxId Γ
	    fromToTypId A
    in refl

  fromToTypId (Γ : Ctx) : (A : Typ Γ) -> FromToTypId A
  fromToTypId {Γ} set	    = rewrite fromToCtxId Γ in refl
  fromToTypId {Γ} (pi A B)  =
    rewrite fromToCtxId Γ
	    fromToTypId A
	    fromToTypId B
    in refl

