
module TypeChecker.Monad.Rewriting where

import Control.Monad.Reader
import Control.Monad.State
import Data.Map as Map

import Syntax.Values

import TypeChecker.Monad
import TypeChecker.Rewriting
import TypeChecker.Abstractions

import Utils.Monads

applyRewriting :: Value -> TC (Either Value Value)
applyRewriting v =
    do	loc  <- envLocalRewriteRules <$> ask
	glob <- tcGlobalRewriteRules <$> get
	r <- applyRewriteRules glob v
	case r of
	    Left v' -> applyRewriteRules loc v'
	    Right w -> return $ Right w

addLocalRewriteRule :: RewriteRule -> TC a -> TC a
addLocalRewriteRule rule =
    local $ \env -> env { envLocalRewriteRules =
			    addRewriteRule rule $ envLocalRewriteRules env
			}

addGlobalRewriteRule :: RewriteRule -> TC ()
addGlobalRewriteRule rule =
    modify $ \st -> st	{ tcGlobalRewriteRules =
			    addRewriteRule rule $ tcGlobalRewriteRules st
			}

raiseLocalRewriteRules :: Int -> TC a -> TC a
raiseLocalRewriteRules k =
    local $ \env -> env { envLocalRewriteRules = raiseBy k
			  $ envLocalRewriteRules env
			}

clearLocalRewriteRules :: TC a -> TC a
clearLocalRewriteRules =
    local $ \env -> env { envLocalRewriteRules = Map.empty }

